# GPTune Copyright (c) 2019, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S.Dept. of Energy) and the University of
# California, Berkeley.  All rights reserved.
#
# If you have questions about your rights to use or distribute this software,
# please contact Berkeley Lab's Intellectual Property Office at IPO@lbl.gov.
#
# NOTICE. This Software was developed under funding from the U.S. Department
# of Energy and the U.S. Government consequently retains certain rights.
# As such, the U.S. Government has been granted for itself and others acting
# on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in
# the Software to reproduce, distribute copies to the public, prepare
# derivative works, and perform publicly and display publicly, and to permit
# other to do so.
#

import abc
from typing import Collection, Tuple
import numpy as np

from problem import Problem
from computer import Computer
from data import Data

import mpi4py

import concurrent
from concurrent import futures
class Model(abc.ABC):

    def __init__(self, problem : Problem, computer : Computer, **kwargs):

        self.problem = problem
        self.computer = computer
        self.M = None

    @abc.abstractmethod
    def train(self, data : Data, **kwargs):

        raise Exception("Abstract method")

    @abc.abstractmethod
    def update(self, newdata : Data, do_train: bool = False, **kwargs):

        raise Exception("Abstract method")

    @abc.abstractmethod
    def predict(self, points : Collection[np.ndarray], tid : int, **kwargs) -> Collection[Tuple[float, float]]:

        raise Exception("Abstract method")


import GPy

class Model_GPy_LCM(Model):

#model_threads=1
#model_processes=1
#model_groups=1
#model_restarts=1
#model_max_iters=15000
#model_latent=0
#model_sparse=False
#model_inducing=None
#model_layers=2

    def train(self, data : Data, **kwargs):

        multitask = len(data.I) > 1

        if (kwargs['model_latent'] is None):
            model_latent = data.NI
        else:
            model_latent = kwargs['model_latent']

        if (kwargs['model_sparse'] and kwargs['model_inducing'] is None):
            if (multitask):
                lenx = sum([len(P) for P in data.P])
            else:
                lenx = len(data.P)
            model_inducing = int(min(lenx, 3 * np.sqrt(lenx)))

        GPy.util.linalg.jitchol.__defaults__ = (kwargs['model_max_jitter_try'],)

        if (multitask):
            kernels_list = [GPy.kern.RBF(input_dim = len(data.P[0][0]), ARD=True) for k in range(model_latent)]
            K = GPy.util.multioutput.LCM(input_dim = len(data.P[0][0]), num_outputs = data.NI, kernels_list = kernels_list, W_rank = 1, name='GPy_LCM')
            K['.*rbf.variance'].constrain_fixed(1.) #For this kernel, K.*.B.kappa and B.W encode the variance now.
            # print(K)
            if (kwargs['model_sparse']):
                self.M = GPy.models.SparseGPCoregionalizedRegression(X_list = data.P, Y_list = data.O, kernel = K, num_inducing = model_inducing)
            else:
                self.M = GPy.models.GPCoregionalizedRegression(X_list = data.P, Y_list = data.O, kernel = K)
            for qq in range(model_latent):
                self.M['.*mixed_noise.Gaussian_noise_%s.variance'%qq].constrain_bounded(1e-10,1e-5)
        else:
            K = GPy.kern.RBF(input_dim = len(data.P[0][0]), ARD=True, name='GPy_GP')
            if (kwargs['model_sparse']):
                self.M = GPy.models.SparseGPRegression(data.P[0], data.O[0], kernel = K, num_inducing = model_inducing)
            else:
                self.M = GPy.models.GPRegression(data.P[0], data.O[0], kernel = K)
            self.M['.*Gaussian_noise.variance'].constrain_bounded(1e-10,1e-5)

#        np.random.seed(mpi_rank)
#        num_restarts = max(1, model_n_restarts // mpi_size)


        resopt = self.M.optimize_restarts(num_restarts = kwargs['model_restarts'], robust = True, verbose = kwargs['verbose'], parallel = (kwargs['model_threads'] > 1), num_processes = kwargs['model_threads'], messages = kwargs['verbose'], optimizer = 'lbfgs', start = None, max_iters = kwargs['model_max_iters'], ipython_notebook = False, clear_after_finish = True)

#        self.M.param_array[:] = allreduce_best(self.M.param_array[:], resopt)[:]
        self.M.parameters_changed()

        if(multitask): # dump the hyperparameters
            theta=[]
            var=[]
            ws=[]
            kappa=[]
            sigma=[]
            # print(self.M)
            for qq in range(model_latent):
                q = self.M.kern['.*GPy_LCM%s.rbf.lengthscale'%qq]
                theta = theta + q.values.tolist()
                q = self.M.kern['.*GPy_LCM%s.rbf.variance'%qq]
                var = var + q.values.tolist() 
                q = self.M.kern['.*GPy_LCM%s.B.W'%qq]
                ws = ws + q.values.tolist() 
                q = self.M.kern['.*GPy_LCM%s.B.kappa'%qq]
                kappa = kappa + q.values.tolist() 
            for qq in range(data.NI):
                q = self.M['.*mixed_noise.Gaussian_noise_%s.variance'%qq]
                sigma = sigma + q.values.tolist()

            if(kwargs['verbose']==True):
                print('theta:',theta)
                print('var:',var)
                print('kappa:',kappa)
                print('WS:',ws)            
                print('sigma:',sigma)



            params = theta+var+kappa+sigma+ws        
        return

    def update(self, newdata : Data, do_train: bool = False, **kwargs):

        #XXX TODO
        self.train(newdata, **kwargs)

    def predict(self, points : Collection[np.ndarray], tid : int, **kwargs) -> Collection[Tuple[float, float]]:

        x = np.empty((1, points.shape[0] + 1))
        x[0,:-1] = points
        x[0,-1] = tid
        (mu, var) = self.M.predict_noiseless(x)

        return (mu, var)

    def get_correlation_metric(self, delta):
        print("In model.py, delta = ", delta)
        Q = delta # number of latent processes 
        B = np.zeros((delta, delta, Q))
        for i in range(Q):
            currentLCM = getattr(self.M.sum, f"GPy_LCM{i}")
            Wq = currentLCM.B.W.values
            Kappa_q = currentLCM.B.kappa.values
            B[:, :, i] = np.outer(Wq, Wq) + np.diag(Kappa_q)
            # print("In model.py, i = ", i)
            # print(B[:, :, i])
            
        # return C_{i, i'}
        C = np.zeros((delta, delta))
        for i in range(delta):
            for ip in range(i, delta):
                C[i, ip] = np.linalg.norm(B[i, ip, :]) / np.sqrt(np.linalg.norm(B[i, i, :]) * np.linalg.norm(B[ip, ip, :]))
        return C

class Model_LCM(Model):

    def train(self, data : Data, **kwargs):

        return self.train_mpi(data, i_am_manager = True, restart_iters=list(range(kwargs['model_restarts'])), **kwargs)

    def train_mpi(self, data : Data, i_am_manager : bool, restart_iters : Collection[int] = None, **kwargs):

        if (kwargs['RCI_mode'] is False):
            from lcm import LCM

        if (kwargs['model_latent'] is None):
            Q = data.NI
        else:
            Q = kwargs['model_latent']

        if (kwargs['distributed_memory_parallelism'] and i_am_manager):
            mpi_comm = self.computer.spawn(__file__, nproc=kwargs['model_restart_processes'], nthreads=kwargs['model_restart_threads'], kwargs=kwargs) # XXX add args and kwargs
            kwargs_tmp = kwargs
            # print("kwargs_tmp",kwargs_tmp)

            if "mpi_comm" in kwargs_tmp:
                del kwargs_tmp["mpi_comm"]   # mpi_comm is not picklable
            _ = mpi_comm.bcast((self, data, restart_iters, kwargs_tmp), root=mpi4py.MPI.ROOT)
            tmpdata = mpi_comm.gather(None, root=mpi4py.MPI.ROOT)
            mpi_comm.Disconnect()
            res=[]
            for p in range(int(kwargs['model_restart_processes'])):
                res = res + tmpdata[p]

        elif (kwargs['shared_memory_parallelism']): #YL: not tested

            #with concurrent.futures.ProcessPoolExecutor(max_workers = kwargs['search_multitask_threads']) as executor:
            with concurrent.futures.ThreadPoolExecutor(max_workers = kwargs['model_restart_threads']) as executor:
                def fun(restart_iter):
                    # if ('seed' in kwargs):
                    #     seed = kwargs['seed'] * kwargs['model_restart_threads'] + restart_iter
                    # else:
                    #     seed = restart_iter
                    # np.random.seed(seed)
                    ## np.random.seed()
                    kern = LCM(input_dim = len(data.P[0][0]), num_outputs = data.NI, Q = Q)
                    # if (restart_iter == 0 and self.M is not None):
                    #     kern.set_param_array(self.M.kern.get_param_array())
                    return kern.train_kernel(X = data.P, Y = data.O, computer = self.computer, kwargs = kwargs)
                res = list(executor.map(fun, restart_iters, timeout=None, chunksize=1))

        else:
            def fun(restart_iter):
                # np.random.seed(restart_iter)
                np.random.seed()
                kern = LCM(input_dim = len(data.P[0][0]), num_outputs = data.NI, Q = Q)
                return kern.train_kernel(X = data.P, Y = data.O, computer = self.computer, kwargs = kwargs)
            res = list(map(fun, restart_iters))

        if (kwargs['distributed_memory_parallelism'] and i_am_manager == False):
            return res

        kern = LCM(input_dim = len(data.P[0][0]), num_outputs = data.NI, Q = Q)
        best_result = min(res, key = lambda x: x[1])
        bestxopt = best_result[0]
        neg_log_marginal_likelihood = best_result[1]
        gradients = best_result[2]
        iteration = best_result[3]
        kern.set_param_array(bestxopt)
        if(kwargs['verbose']==True):
            # print('hyperparameters:', kern.get_param_array())
            print('theta:',kern.theta)
            print('var:',kern.var)
            print('kappa:',kern.kappa)
            print('sigma:',kern.sigma)
            print('WS:',kern.WS)




        # YL: likelihoods needs to be provided, since K operator doesn't take into account sigma/jittering, but Kinv does. The GPCoregionalizedRegression intialization will call inference in GPy/interence/latent_function_inference/exact_gaussian_inference.py, and add to diagonals of the K operator with sigma+1e-8
        likelihoods_list = [GPy.likelihoods.Gaussian(variance = kern.sigma[i], name = "Gaussian_noise_%s" %i) for i in range(data.NI)]
        self.M = GPy.models.GPCoregionalizedRegression(data.P, data.O, kern, likelihoods_list = likelihoods_list)

        #print ("kernel: " + str(kern))
        #print ("bestxopt:" + str(bestxopt))
        #print ("neg_log_marginal_likelihood:" + str(neg_log_marginal_likelihood))
        #print ("gradients: " + str(gradients))
        #print ("iteration: " + str(iteration))
        #for i in range(data.NI):
        #    print ("i: " + str(i))
        #    print ("sigma: " + str(kern.sigma[i]))
        #    print ("likelihoods_list: " + str(likelihoods_list[i].to_dict()))
        #print ("likelihoods_list_len: " + str(data.NI))
        #print ("self.M: " + str(self.M))

        return (bestxopt, neg_log_marginal_likelihood, gradients, iteration)

    def update(self, newdata : Data, do_train: bool = False, **kwargs):

        #XXX TODO
        self.train(newdata, **kwargs)

    # make prediction on a single sample point of a specific task tid
    def predict(self, points : Collection[np.ndarray], tid : int, **kwargs) -> Collection[Tuple[float, float]]:

        x = np.empty((1, points.shape[0] + 1))
        x[0,:-1] = points
        x[0,-1] = tid
        (mu, var) = self.M.predict_noiseless(x)   # predict_noiseless ueses precomputed Kinv and Kinv*y (generated at GPCoregionalizedRegression init, which calls inference in GPy/inference/latent_function_inference/exact_gaussian_inference.py) to compute mu and var, with O(N^2) complexity, see "class PosteriorExact(Posterior): _raw_predict" of GPy/inference/latent_function_inference/posterior.py.

        return (mu, var)

    def gen_model_from_hyperparameters(self, data : Data, hyperparameters : list, **kwargs):
        if (kwargs['RCI_mode'] is False):
            from lcm import LCM

        if (kwargs['model_latent'] is None):
            Q = data.NI
        else:
            Q = kwargs['model_latent']

        kern = LCM(input_dim = len(data.P[0][0]), num_outputs = data.NI, Q = Q)
        #print ("received hyperparameters: " + str(hyperparameters))
        kern.set_param_array(hyperparameters)

        likelihoods_list = [GPy.likelihoods.Gaussian(variance = kern.sigma[i], name = "Gaussian_noise_%s" %i) for i in range(data.NI)]
        self.M = GPy.models.GPCoregionalizedRegression(data.P, data.O, kern, likelihoods_list = likelihoods_list)

        return

class Model_DGP(Model):

    def train(self, data : Data, **kwargs):

        multitask = len(self.I) > 1

        if (multitask):
            X = np.array([np.concatenate((self.I[i], self.P[i][j])) for i in range(len(self.I)) for j in range(self.P[i].shape[0])])
        else:
            X = self.P[0]
        Y = np.array(list(itertools.chain.from_iterable(self.O)))

        #--------- Model Construction ----------#
        model_n_layers = 2
        # Define what kernels to use per layer
        kerns = [GPy.kern.RBF(input_dim=Q, ARD=True) + GPy.kern.Bias(input_dim=Q) for lev in range(model_n_layers)]
        kerns.append(GPy.kern.RBF(input_dim=X.shape[1], ARD=True) + GPy.kern.Bias(input_dim=X.shape[1]))
        # Number of inducing points to use
        if (num_inducing is None):
            if (multitask):
                lenx = sum([len(X) for X in self.P])
            else:
                lenx = len(self.P)
#            num_inducing = int(min(lenx, 3 * np.sqrt(lenx)))
            num_inducing = lenx
        # Whether to use back-constraint for variational posterior
        back_constraint = False
        # Dimensions of the MLP back-constraint if set to true
        encoder_dims=[[X.shape[0]],[X.shape[0]],[X.shape[0]]]

        nDims = [Y.shape[1]] + model_n_layers * [Q] + [X.shape[1]]
#        self.M = deepgp.DeepGP(nDims, Y, X=X, num_inducing=num_inducing, likelihood = None, inits='PCA', name='deepgp', kernels=kerns, obs_data='cont', back_constraint=True, encoder_dims=encoder_dims, mpi_comm=mpi_comm, self.mpi_root=0, repeatX=False, inference_method=None)#, **kwargs):
        self.M = deepgp.DeepGP(nDims, Y, X=X, num_inducing=num_inducing, likelihood = None, inits='PCA', name='deepgp', kernels=kerns, obs_data='cont', back_constraint=False, encoder_dims=None, mpi_comm=None, mpi_root=0, repeatX=False, inference_method=None)#, **kwargs):
#        self.M = deepgp.DeepGP([Y.shape[1], Q, Q, X.shape[1]], Y, X=X, kernels=[kern1, kern2, kern3], num_inducing=num_inducing, back_constraint=back_constraint)

        #--------- Optimization ----------#
        # Make sure initial noise variance gives a reasonable signal to noise ratio.
        # Fix to that value for a few iterations to avoid early local minima
        for i in range(len(self.M.layers)):
            output_var = self.M.layers[i].Y.var() if i==0 else self.M.layers[i].Y.mean.var()
            self.M.layers[i].Gaussian_noise.variance = output_var*0.01
            self.M.layers[i].Gaussian_noise.variance.fix()

        self.M.optimize_restarts(num_restarts = num_restarts, robust = True, verbose = self.verbose, parallel = (num_processes is not None), num_processes = num_processes, messages = "True", optimizer = 'lbfgs', start = None, max_iters = max_iters, ipython_notebook = False, clear_after_finish = True)

        # Unfix noise variance now that we have initialized the model
        for i in range(len(self.M.layers)):
            self.M.layers[i].Gaussian_noise.variance.unfix()

        self.M.optimize_restarts(num_restarts = num_restarts, robust = True, verbose = self.verbose, parallel = (num_processes is not None), num_processes = num_processes, messages = "True", optimizer = 'lbfgs', start = None, max_iters = max_iters, ipython_notebook = False, clear_after_finish = True)

    def update(self, newdata : Data, do_train: bool = False, **kwargs):

        #XXX TODO
        self.train(newdata, **kwargs)

    def predict(self, points : Collection[np.ndarray], tid : int, **kwargs) -> Collection[Tuple[float, float]]:

        (mu, var) = self.M.predict(np.concatenate((self.I[tid], x)).reshape((1, self.DT + self.DI)))

        return (mu, var)


if __name__ == '__main__':
    def objectives(point):
        print('this is a dummy definition')
        return point
    def models(point):
        print('this is a dummy definition')
        return point
    def models_update(data):
        print('this is a dummy definition')
        return data        
    def cst1(point):
        print('this is a dummy definition')
        return point
    def cst2(point):
        print('this is a dummy definition')
        return point
    def cst3(point):
        print('this is a dummy definition')
        return point
    def cst4(point):
        print('this is a dummy definition')
        return point
    def cst5(point):
        print('this is a dummy definition')
        return point                        
    from mpi4py import MPI
    mpi_comm = mpi4py.MPI.Comm.Get_parent()
    mpi_rank = mpi_comm.Get_rank()
    mpi_size = mpi_comm.Get_size()
    (modeler, data, restart_iters, kwargs) = mpi_comm.bcast(None, root=0)
    restart_iters_loc = restart_iters[mpi_rank:len(restart_iters):mpi_size]
    tmpdata = modeler.train_mpi(data, i_am_manager = False, restart_iters = restart_iters_loc, **kwargs)
    res = mpi_comm.gather(tmpdata, root=0)
    mpi_comm.Disconnect()




# GPTune Copyright (c) 2019, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S.Dept. of Energy) and the University of
# California, Berkeley.  All rights reserved.
#
# If you have questions about your rights to use or distribute this software,
# please contact Berkeley Lab's Intellectual Property Office at IPO@lbl.gov.
#
# NOTICE. This Software was developed under funding from the U.S. Department
# of Energy and the U.S. Government consequently retains certain rights.
# As such, the U.S. Government has been granted for itself and others acting
# on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in
# the Software to reproduce, distribute copies to the public, prepare
# derivative works, and perform publicly and display publicly, and to permit
# other to do so.
#

import copy
import functools
import time

from autotune.problem import TuningProblem
from autotune.space import *

from problem import Problem
from computer import Computer
from options import Options
from data import *
from database import *
from sample import *
from model import *
from search import *
import math
import os

import mpi4py
import numpy as np

import json
from filelock import Timeout, FileLock

class GPTune(object):

    def __init__(self, tuningproblem : TuningProblem, computer : Computer = None, data : Data = None, historydb : HistoryDB = None, options : Options = None, driverabspath=None, models_update=None, **kwargs):

        """
        tuningproblem: object defining the characteristics of the tuning (See file 'autotuner/autotuner/tuningproblem.py')
        computer     : object specifying the architectural characteristics of the computer to run on (See file 'GPTune/computer.py')
        data         : object containing the data of a previous tuning (See file 'GPTune/data.py')
        options      : object defining all the options that will define the behaviour of the tuner (See file 'GPTune/options.py')
        """
        self.tuningproblem = tuningproblem
        self.problem  = Problem(tuningproblem,driverabspath=driverabspath,models_update=models_update)
        if (computer is None):
            computer = Computer()
        self.computer = computer
        if (data is None):
            data = Data(self.problem)
        self.data     = data
        if (options is None):
            options = Options()
        self.options  = options
        if (historydb is None):
            historydb = HistoryDB()
        self.historydb = historydb

    def LoadSurrogateModel(self, model_data : dict, **kwargs):

        Igiven = model_data["task_parameters"]

        """ Load history function evaluation data """

        self.historydb.load_model_func_eval(data = self.data, problem = self.problem,
                Igiven = Igiven, model_data = model_data)

        """ Update data space """

        if(Igiven is not None and self.data.I is None):
            self.data.I = Igiven

        # normalize the data as the user always work in the original space
        if self.data.P is not None:
            tmp=[]
            for x in self.data.P:
                xNorm = self.problem.PS.transform(x)
                tmp.append(xNorm)
            self.data.P=tmp
        if self.data.I is not None:
            self.data.I = self.problem.IS.transform(self.data.I)

        if (self.data.O is None and self.data.P is not None and self.data.I is not None):
            self.data.O = self.computer.evaluate_objective(self.problem, self.data.I, self.data.P, self.data.D, options = kwargs)

        if (self.data.D is None):
            self.data.D = [{}] * len(self.data.I)

        if (self.data.P is not None and len(self.data.P) !=len(self.data.I)):
            raise Exception("len(self.data.P) !=len(self.data.I)")

        if (self.data.O is not None and len(self.data.O) !=len(self.data.I)):
            raise Exception("len(self.data.O) !=len(self.data.I)")

        options1 = copy.deepcopy(self.options)
        kwargs.update(options1)

        """ Reproduce surrogate models """

        modelers = [eval(f'{kwargs["model_class"]} (problem = self.problem, computer = self.computer)')]*self.problem.DO

        for i in range(self.problem.DO):
            modelers[i].gen_model_from_hyperparameters(self.data,
                    model_data["hyperparameters"],
                    **kwargs)

        def model_function(point):

            task_parameter_names = [self.problem.IS[k].name for k in range(len(self.problem.IS))]
            tuning_parameter_names = [self.problem.PS[k].name for k in range(len(self.problem.PS))]
            tuning_parameter_types = [type(self.problem.PS[k]).__name__ for k in range(len(self.problem.PS))]
            output_names = [self.problem.OS[k].name for k in range(len(self.problem.OS))]

            #print ("task_parameter_names: ", task_parameter_names)
            #print ("tuning_parameter_names: ", tuning_parameter_names)
            #print ("tuning_parameter_types: ", tuning_parameter_types)
            #print ("output_names: ", output_names)
            #print ("Igiven: ", Igiven)

            input_task = []
            for task_parameter_name in task_parameter_names:
                input_task.append(point[task_parameter_name])

            tid = -1
            for i in range(len(Igiven)):
                is_equal = True
                for j in range(len(Igiven[i])):
                    if Igiven[i][j] != input_task[j]:
                        is_equal = False
                if (is_equal):
                    tid = i
                    break
            if tid == -1:
                print ("[Error] cannot find model for the given input task: ", input_task)
                return None

            bound_checked = True

            input_tuning_parameters = []
            for tuning_parameter_name in tuning_parameter_names:
                if tuning_parameter_types[tuning_parameter_names.index(tuning_parameter_name)] == "Integer" or\
                   tuning_parameter_types[tuning_parameter_names.index(tuning_parameter_name)] == "Integer":
                    lower_bound, upper_bound = self.problem.PS.bounds[tuning_parameter_names.index(tuning_parameter_name)]
                    if point[tuning_parameter_name] < lower_bound or point[tuning_parameter_name] > upper_bound:
                        bound_checked = False
                input_tuning_parameters.append(point[tuning_parameter_name])

            #tuning_parameter_names_model_order = parameter_names
            #print ("tuning_parameter_names_model_order")
            #print (tuning_parameter_names_model_order)
            #input_tuning_parameters_transformed_reordered = []
            #for parameter_name in tuning_parameter_names_model_order:
            #    print ("parameter_name: ", parameter_name)
            #    print ("index: ", tuning_parameter_names.index(parameter_name))
            #    print ("val: ", input_tuning_parameters_transformed[tuning_parameter_names.index(parameter_name)])
            #    input_tuning_parameters_transformed_reordered.append(
            #            input_tuning_parameters_transformed[tuning_parameter_names.index(parameter_name)])
            #print ("input_tuning_parameters_transformed_reordered")
            #print (input_tuning_parameters_transformed_reordered)

            ret = {}

            if (bound_checked == True):
                input_tuning_parameters_transformed = self.problem.PS.transform([input_tuning_parameters])[0]

                for o in range(self.problem.DO):
                    (mu, var) = modelers[o].predict(np.array(input_tuning_parameters_transformed),tid)

                    ret[output_names[o]] = np.array(mu).tolist()
                    ret[output_names[o]+"_var"] = var
            else:
                for o in range(self.problem.DO):
                    # TODO: the value depends on the problem (output) definition
                    (mu, var) = [[1000000.0]], 0
                    ret[output_names[o]] = np.array(mu).tolist()
                    ret[output_names[o]+"_var"] = var

            ret["source"] = "model_function"

            return ret

        return (modelers, model_function)

    def MLA_LoadModel(self, NS = 0, Igiven = None, method = "max_evals", update = 0, model_uids = None, **kwargs):
        print('\n\n\n------Starting MLA with Trained Model for %d tasks and %d samples each '%(len(Igiven),NS))
        stats = {
            "time_total": 0,
            "time_fun": 0,
            "time_search": 0,
            "time_model": 0,
            "modeling_time":[],
            "modeling_iteration":[]
        }
        time_fun=0
        time_search=0
        time_model=0

        """ Load history function evaluation data """
        self.historydb.load_history_func_eval(self.data, self.problem, Igiven)
        np.set_printoptions(suppress=False,precision=4)
        NSmin=0
        if (self.data.P is not None):
            NSmin = min(map(len, self.data.P)) # the number of samples per task in existing tuning data can be different

        #if (self.data.P is not None and NSmin>=NS and self.data.O is not None):
        #    print('NSmin>=NS, no need to run MLA. Returning...')
        #    return (copy.deepcopy(self.data), None,stats)
        """ Set (additional) number of samples for autotuning """
        NS = NSmin + NS

        t3 = time.time_ns()

        options1 = copy.deepcopy(self.options)
        kwargs.update(options1)

        """ Multi-task Learning Autotuning """
        if(Igiven is not None and self.data.I is None):  # building the MLA model for each of the given tasks
            self.data.I = Igiven

        # normalize the data as the user always work in the original space
        if self.data.P is not None: # from a list of (list of lists) to a list of 2D numpy arrays
            tmp=[]
            for x in self.data.P:
                xNorm = self.problem.PS.transform(x)
                tmp.append(xNorm)
            self.data.P=tmp
        if self.data.I is not None: # from a list of lists to a 2D numpy array
            self.data.I = self.problem.IS.transform(self.data.I)

        if (self.data.O is None and self.data.P is not None and self.data.I is not None): # tuning parameters and task parameters are given, but the output is none
            self.data.O = self.computer.evaluate_objective(self.problem, self.data.I, self.data.P, self.data.D, options = kwargs)

        if (self.data.D is None):
            self.data.D = [{}] * len(self.data.I) #NI

        if (self.data.P is not None and len(self.data.P) !=len(self.data.I)):
            raise Exception("len(self.data.P) !=len(self.data.I)")

        if (self.data.O is not None and len(self.data.O) !=len(self.data.I)):
            raise Exception("len(self.data.O) !=len(self.data.I)")

        modelers  = [eval(f'{kwargs["model_class"]} (problem = self.problem, computer = self.computer)')]*self.problem.DO
        for i in range(self.problem.DO):
            # current limitations
            # - only model LCM
            # - not considering edge cases (e.g. no model is available)
            if model_uids == None:
                #TODO CHECK: make self.data is correct (we may need to load (or double check) func eval data based on the model data)

                if method == "max_evals":
                    (hyperparameters, parameter_names) = self.historydb.load_max_evals_surrogate_model_hyperparameters(
                            self.tuningproblem, Igiven, i, kwargs["model_class"])
                elif method == "MLE" or method == "mle":
                    (hyperparameters, parameter_names) = self.historydb.load_MLE_surrogate_model_hyperparameters(
                            self.tuningproblem, Igiven, i, kwargs["model_class"])
                elif method == "AIC" or method == "aic":
                    (hyperparameters, parameter_names) = self.historydb.load_AIC_surrogate_model_hyperparameters(
                            self.tuningproblem, Igiven, i, kwargs["model_class"])
                elif method == "BIC" or method == "bic":
                    (hyperparameters, parameter_names) = self.historydb.load_BIC_model_hyperparameters(
                            self.tuningproblem, Igiven, i, kwargs["model_class"])
                else:
                    (hyperparameters, parameter_names) = self.historydb.load_max_evals_surrogate_model_hyperparameters(
                            self.tuningproblem, Igiven, i, kwargs["model_class"])

            else:
                (hyperparameters, parameter_names) = self.historydb.load_surrogate_model_hyperparameters_by_uid(model_uids[i])
            modelers[i].gen_model_from_hyperparameters(self.data,
                    hyperparameters,
                    **kwargs)

        searcher = eval(f'{kwargs["search_class"]}(problem = self.problem, computer = self.computer, options = self.options)')
        model_reupdate = 0
        if update == -1: # search one sample without updating model; then search next with updating model.
            model_reupdate = -1
            update = 1
        optiter = 0
        NSmin = min(map(len, self.data.P))
        while NSmin<NS:# YL: each iteration adds 1 (if single objective) or at most kwargs["search_more_samples"] (if multi-objective) sample until total #sample reaches NS

            if(self.problem.models_update is not None):
                ########## denormalize the data as the user always work in the original space
                tmpdata = copy.deepcopy(self.data)
                if tmpdata.I is not None:    # from 2D numpy array to a list of lists
                    tmpdata.I = self.problem.IS.inverse_transform(tmpdata.I)
                if tmpdata.P is not None:    # from a collection of 2D numpy arrays to a list of (list of lists)
                    tmpdata.P = [self.problem.PS.inverse_transform(x) for x in tmpdata.P]
                self.problem.models_update(tmpdata)
                self.data.D = tmpdata.D

            newdata = Data(problem = self.problem, I = self.data.I, D = self.data.D)
            print("MLA iteration: ",optiter)
            stats["modeling_iteration"].append(0)
            optiter = optiter + 1
            model_reupdate = model_reupdate + 1
            t1 = time.time_ns()
            for o in range(self.problem.DO):
                tmpdata = copy.deepcopy(self.data)
                tmpdata.O = [copy.deepcopy(self.data.O[i][:,o].reshape((-1,1))) for i in range(len(self.data.I))]
                if(self.problem.models is not None):
                    for i in range(len(tmpdata.P)):
                        points0 = tmpdata.D[i]
                        t = tmpdata.I[i]
                        I_orig = self.problem.IS.inverse_transform(np.array(t, ndmin=2))[0]
                        points1 = {self.problem.IS[k].name: I_orig[k] for k in range(self.problem.DI)}
                        modeldata=[]
                        for p in range(len(tmpdata.P[i])):
                            x = tmpdata.P[i][p]
                            x_orig = self.problem.PS.inverse_transform(np.array(x, ndmin=2))[0]
                            points = {self.problem.PS[k].name: x_orig[k] for k in range(self.problem.DP)}
                            points.update(points1)
                            points.update(points0)
                            modeldata.append(self.problem.models(points))
                        modeldata=np.array(modeldata)
                        tmpdata.P[i] = np.hstack((tmpdata.P[i],modeldata))  # YL: here tmpdata in the normalized space, but modeldata is the in the original space
                for i in range(len(tmpdata.P)):   # LCM requires the same number of samples per task, so use the first NSmin samples
                    tmpdata.O[i] = tmpdata.O[i][0:NSmin,:]
                    tmpdata.P[i] = tmpdata.P[i][0:NSmin,:]

                if model_reupdate == update:
                    # print(tmpdata.P[0])
                    #print ("[bestxopt]: len: " + str(len(bestxopt)) + " val: " + str(bestxopt))
                    if (kwargs["model_class"] == "Model_LCM"):
                        (bestxopt, neg_log_marginal_likelihood,
                                gradients, iteration) = \
                            modelers[o].train(data = tmpdata, **kwargs)
                        self.historydb.store_model_LCM(
                                o,
                                self.problem,
                                self.data.I,
                                bestxopt,
                                neg_log_marginal_likelihood,
                                gradients,
                                iteration)
                        stats["modeling_iteration"][optiter-1] += iteration
                    else:
                        modelers[o].train(data = tmpdata, **kwargs)
                        stats["modeling_iteration"][optiter-1] += 0
                    model_reupdate = 0
                else:
                    if (kwargs["model_class"] == "Model_LCM"):
                        stats["modeling_iteration"][optiter-1] += 0

            t2 = time.time_ns()
            stats["modeling_time"].append((t2-t1)/1e9)
            time_model = time_model + (t2-t1)/1e9

            t1 = time.time_ns()
            res = searcher.search_multitask(data = self.data, models = modelers, **kwargs)

            newdata.P = [x[1][0] for x in res]
            for i in range(len(newdata.P)):  # if NSi>=NS, skip the function evaluation
                NSi = self.data.P[i].shape[0]
                newdata.P[i] = newdata.P[i][0:min(newdata.P[i].shape[0],max(0,NS-NSi)),:]
            # print(more_samples,newdata.P)
            t2 = time.time_ns()
            time_search = time_search + (t2-t1)/1e9

            t1 = time.time_ns()
            newdata.O = self.computer.evaluate_objective(problem = self.problem,
                    I = newdata.I,
                    P = newdata.P,
                    D = newdata.D,
                    history_db = self.historydb,
                    options = kwargs)
            t2 = time.time_ns()
            time_fun = time_fun + (t2-t1)/1e9
            self.data.merge(newdata)
            NSmin = min(map(len, self.data.P))

        # denormalize the data as the user always work in the original space
        if self.data.I is not None:    # from 2D numpy array to a list of lists
            self.data.I = self.problem.IS.inverse_transform(self.data.I)
        if self.data.P is not None:    # from a collection of 2D numpy arrays to a list of (list of lists)
            self.data.P = [self.problem.PS.inverse_transform(x) for x in self.data.P]

        t4 = time.time_ns()
        stats['time_total'] = (t4-t3)/1e9
        stats['time_fun'] = time_fun
        stats['time_model'] = time_model
        stats['time_search'] = time_search

        return (copy.deepcopy(self.data), modelers, stats)

    def MLA_HistoryDB(self, NS, NS1 = None, NI = None, Igiven = None, **kwargs):
        print('\n\n\n------Starting MLA with HistoryDB with %d tasks and %d samples each '%(NI,NS))
        stats = {
            "time_total": 0,
            "time_sample_init": 0,
            "time_fun": 0,
            "time_search": 0,
            "time_model": 0,
            "func_eval_time":[],
            "search_time":[],
            "modeling_time":[],
            "modeling_iteration":[]
        }
        time_fun=0
        time_sample_init=0
        time_search=0
        time_model=0

        """ Load history function evaluation data """
        if self.historydb.load_func_eval == True:
            self.historydb.load_history_func_eval(self.data, self.problem, Igiven)

        np.set_printoptions(suppress=False,precision=4)
        NSmin=0
        NSmax=0
        if (self.data.P is not None):
            NSmin = min(map(len, self.data.P)) # the number of samples per task in existing tuning data can be different
            NSmax = max(map(len, self.data.P)) 

        # """ Set (additional) number of samples for autotuning """
        # NS = NSmin + NS
        if(NSmax>0):
            if (self.data.P is not None and NSmin>=NS and self.data.O is not None):
                print('\nexisting data has at least NSmin=%d samples per task, which is no less than NS=%d, no need to run MLA. Returning...\n'%(NSmin,NS))
                return (copy.deepcopy(self.data), None,stats)
            else:            
                print('\nexisting data has at least NSmin=%d samples per task, GPTune will generate at most NS-NSmin=%d additional samples.\n'%(NSmin,NS-NSmin))


        t3 = time.time_ns()

        t1 = time.time_ns()

        options1 = copy.deepcopy(self.options)
        kwargs.update(options1)

        """ Multi-task Learning Autotuning """

        if (NS1 is not None and NS1>NS):
            raise Exception("NS1>NS")
        if (NS1 is None):
            NS1 = min(NS - 1, 3 * self.problem.DP) # heuristic rule in literature

        if(Igiven is not None and self.data.I is None):  # building the MLA model for each of the given tasks
            self.data.I = Igiven

########## normalize the data as the user always work in the original space
        if self.data.P is not None: # from a list of (list of lists) to a list of 2D numpy arrays
            tmp=[]
            for x in self.data.P:
                if(len(x)>0):
                    xNorm = self.problem.PS.transform(x)
                    tmp.append(xNorm)
                else:
                    tmp.append(np.empty( shape=(0, self.problem.DP) ))
            self.data.P=tmp
        if self.data.I is not None: # from a list of lists to a 2D numpy array
            self.data.I = self.problem.IS.transform(self.data.I)

        if (self.data.O is None and self.data.P is not None and self.data.I is not None): # tuning parameters and task parameters are given, but the output is none
            self.data.O = self.computer.evaluate_objective(self.problem, self.data.I, self.data.P, self.data.D, self.historydb, options = kwargs)

        sampler = eval(f'{kwargs["sample_class"]}()')
        if (self.data.I is None):

            if (NI is None):
                raise Exception("Number of problems to be generated (NI) is not defined")

            check_constraints = functools.partial(self.computer.evaluate_constraints, self.problem, inputs_only = True, kwargs = kwargs)
            self.data.I = sampler.sample_inputs(n_samples = NI, IS = self.problem.IS, check_constraints = check_constraints, **kwargs)
            # print("riji",type(self.data.I),type(self.data.I[0]))
            self.data.D = [{}] * NI
        else:
            if (self.data.D is None):
                self.data.D = [{}] * NI

        if (self.data.P is not None and len(self.data.P) !=len(self.data.I)):
            raise Exception("len(self.data.P) !=len(self.data.I)")

        if (NSmin<NS1):
            check_constraints = functools.partial(self.computer.evaluate_constraints, self.problem, inputs_only = False, kwargs = kwargs)
            tmpP = sampler.sample_parameters(n_samples = NS1-NSmin, I = self.data.I, IS = self.problem.IS, PS = self.problem.PS, check_constraints = check_constraints, **kwargs)
            if(self.data.P is not None):
                for i in range(len(self.data.P)):
                    NSi = self.data.P[i].shape[0]
                    tmpP[i] = tmpP[i][0:max(NS1-NSi,0),:] # if NSi>=NS1, no need to generate new random data

        if (self.data.O is not None and len(self.data.O) !=len(self.data.I)):
            raise Exception("len(self.data.O) !=len(self.data.I)")

        t2 = time.time_ns()
        time_sample_init = time_sample_init + (t2-t1)/1e9

        t1 = time.time_ns()
        if (NSmin<NS1):
            tmpO = self.computer.evaluate_objective(self.problem, self.data.I, tmpP, self.data.D, self.historydb, options = kwargs)
            if(self.data.P is None): # no existing tuning data is available
                self.data.O = tmpO
                self.data.P = tmpP
            else:                
                for i in range(len(self.data.P)):
                    self.data.P[i] = np.vstack((self.data.P[i],tmpP[i]))
                    self.data.O[i] = np.vstack((self.data.O[i],tmpO[i]))

        t2 = time.time_ns()
        stats["func_eval_time"].append((t2-t1)/1e9)
        time_fun = time_fun + (t2-t1)/1e9

        modelers  = [eval(f'{kwargs["model_class"]} (problem = self.problem, computer = self.computer)')]*self.problem.DO
        searcher = eval(f'{kwargs["search_class"]}(problem = self.problem, computer = self.computer, options = self.options)')
        optiter = 0
        NSmin = min(map(len, self.data.P))
        while NSmin<NS:# YL: each iteration adds 1 (if single objective) or at most kwargs["search_more_samples"] (if multi-objective) sample until total #sample reaches NS

            if(self.problem.models_update is not None):
                ########## denormalize the data as the user always work in the original space
                tmpdata = copy.deepcopy(self.data)
                if tmpdata.I is not None:    # from 2D numpy array to a list of lists
                    tmpdata.I = self.problem.IS.inverse_transform(tmpdata.I)
                if tmpdata.P is not None:    # from a collection of 2D numpy arrays to a list of (list of lists)
                    tmpdata.P = [self.problem.PS.inverse_transform(x) for x in tmpdata.P]
                self.problem.models_update(tmpdata)
                self.data.D = tmpdata.D

            newdata = Data(problem = self.problem, I = self.data.I, D = self.data.D)
            print("MLA iteration: ",optiter)
            stats["modeling_iteration"].append(0)
            optiter = optiter + 1
            t1 = time.time_ns()
            for o in range(self.problem.DO):

                tmpdata = copy.deepcopy(self.data)
                tmpdata.O = [copy.deepcopy(self.data.O[i][:,o].reshape((-1,1))) for i in range(len(self.data.I))]
                if(self.problem.models is not None):
                    for i in range(len(tmpdata.P)):
                        points0 = tmpdata.D[i]
                        t = tmpdata.I[i]
                        I_orig = self.problem.IS.inverse_transform(np.array(t, ndmin=2))[0]
                        points1 = {self.problem.IS[k].name: I_orig[k] for k in range(self.problem.DI)}
                        modeldata=[]
                        for p in range(len(tmpdata.P[i])):
                            x = tmpdata.P[i][p]
                            x_orig = self.problem.PS.inverse_transform(np.array(x, ndmin=2))[0]
                            points = {self.problem.PS[k].name: x_orig[k] for k in range(self.problem.DP)}
                            points.update(points1)
                            points.update(points0)
                            if(self.problem.constants is not None):
                                points.update(self.problem.constants)
                            modeldata.append(self.problem.models(points))
                        modeldata=np.array(modeldata)
                        tmpdata.P[i] = np.hstack((tmpdata.P[i],modeldata))  # YL: here tmpdata in the normalized space, but modeldata is the in the original space
                for i in range(len(tmpdata.P)):   # LCM requires the same number of samples per task, so use the first NSmin samples
                    tmpdata.O[i] = tmpdata.O[i][0:NSmin,:]
                    tmpdata.P[i] = tmpdata.P[i][0:NSmin,:]
                # print(tmpdata.P[0])
                #print ("[bestxopt]: len: " + str(len(bestxopt)) + " val: " + str(bestxopt))
                if (kwargs["model_class"] == "Model_LCM"):
                    if (kwargs["model_output_constraint"] == True):
                        for i in range(len(tmpdata.O[0])):
                            out_of_range = False
                            for o_ in range(self.problem.DO):
                                output_space = self.historydb.problem_space_to_dict(self.problem.OS)[o_]
                                lower_bound = output_space["lower_bound"]
                                upper_bound = output_space["upper_bound"]
                                output_result = [copy.deepcopy(self.data.O[i][:,o_].reshape((-1,1))) for i in range(len(self.data.I))][0][i]
                                if output_result < lower_bound or \
                                   output_result > upper_bound:
                                    out_of_range = True

                            if out_of_range == True or self.historydb.problem_space_to_dict(self.problem.OS)[o]["optimize"] == False:
                                tmpdata.O[0][i][0] = 1000000000.0 #sys.float_info.max

                    (bestxopt, neg_log_marginal_likelihood,
                            gradients, iteration) = \
                        modelers[o].train(data = tmpdata, **kwargs)
                    self.historydb.store_model_LCM(
                            o,
                            self.problem,
                            self.data.I,
                            bestxopt,
                            neg_log_marginal_likelihood,
                            gradients,
                            iteration)
                    stats["modeling_iteration"][optiter-1] += iteration
                else:
                    modelers[o].train(data = tmpdata, **kwargs)
                
                if self.options['verbose'] == True and self.options['model_class'] == 'Model_LCM' and len(self.data.I)>1:
                    C = modelers[o].M.kern.get_correlation_metric()
                    print("The correlation matrix C is \n", C)
                elif self.options['verbose'] == True and self.options['model_class'] == 'Model_GPy_LCM' and len(self.data.I)>1:
                    C = modelers[o].get_correlation_metric(len(self.data.I))
                    print("The correlation matrix C is \n", C)

            t2 = time.time_ns()
            stats["modeling_time"].append((t2-t1)/1e9)
            time_model = time_model + (t2-t1)/1e9

            t1 = time.time_ns()
            res = searcher.search_multitask(data = self.data, models = modelers, **kwargs)

            newdata.P = [x[1][0] for x in res]
            for i in range(len(newdata.P)):  # if NSi>=NS, skip the function evaluation
                NSi = self.data.P[i].shape[0]
                newdata.P[i] = newdata.P[i][0:min(newdata.P[i].shape[0],max(0,NS-NSi)),:]
            # print(more_samples,newdata.P)
            t2 = time.time_ns()
            stats["search_time"].append((t2-t1)/1e9)
            time_search = time_search + (t2-t1)/1e9

            t1 = time.time_ns()
            newdata.O = self.computer.evaluate_objective(problem = self.problem,
                    I = newdata.I,
                    P = newdata.P,
                    D = newdata.D,
                    history_db = self.historydb,
                    options = kwargs)
            t2 = time.time_ns()
            time_fun = time_fun + (t2-t1)/1e9
            self.data.merge(newdata)
            NSmin = min(map(len, self.data.P))

        # denormalize the data as the user always work in the original space
        if self.data.I is not None:    # from 2D numpy array to a list of lists
            self.data.I = self.problem.IS.inverse_transform(self.data.I)
        if self.data.P is not None:    # from a collection of 2D numpy arrays to a list of (list of lists)
            self.data.P = [self.problem.PS.inverse_transform(x) for x in self.data.P]

        t4 = time.time_ns()
        stats['time_total'] = (t4-t3)/1e9
        stats['time_fun'] = time_fun
        stats['time_model'] = time_model
        stats['time_search'] = time_search
        stats['time_sample_init'] = time_sample_init

        return (copy.deepcopy(self.data), modelers, stats)

    def MLA(self, NS, NS1 = None, NI = None, Igiven = None, **kwargs):
        if self.historydb.history_db is True:
            if self.historydb.load_func_eval == True and self.historydb.load_surrogate_model == True:
                return self.MLA_LoadModel(NS = NS, Igiven = Igiven)
            else:
                return self.MLA_HistoryDB(NS, NS1, NI, Igiven)

    def TLA1(self, Tnew, NS):

        print('\n\n\n------Starting TLA1 for task: ',Tnew)

        stats = {
            "time_total": 0,
            "time_fun": 0
        }
        time_fun=0

        t3=time.time_ns()
        # Initialization
        kwargs = copy.deepcopy(self.options)
        ntso = len(self.data.I)
        ntsn = len(Tnew)

        if(self.data.O[0].shape[1]>1):
            raise Exception("TLA1 only works for single-objective tuning")

        PSopt =[]
        for i in range(ntso):
            PSopt.append(self.data.P[i][np.argmin(self.data.O[i])])
        # YSopt = np.array([[self.data.O[k].min()] for k in range(ntso)])
        MSopt = []



        # convert the task spaces to the normalized spaces
        INorms=[]
        for t in self.data.I:
            INorm = self.problem.IS.transform(np.array(t, ndmin=2))[0]
            INorms.append(INorm.reshape((-1, self.problem.DI)))
        INorms = np.vstack([INorms[i] for i in range(ntso)]).reshape((ntso,self.problem.DI))

        tmp=[]
        for t in Tnew:
            INorm = self.problem.IS.transform(np.array(t, ndmin=2))[0]
            tmp.append(INorm.reshape((-1, self.problem.DI)))
        InewNorms=np.vstack([tmp[i] for i in range(ntsn)]).reshape((ntsn,self.problem.DI))

        # convert the parameter spaces to the normalized spaces
        PSoptNorms = self.problem.PS.transform(PSopt)
        columns = []
        for j in range(self.problem.DP):
            columns.append([])
        for i in range(ntso):
            for j in range(self.problem.DP):
                columns[j].append(PSoptNorms[i][j])
        PSoptNorms = []
        for j in range(self.problem.DP):
            PSoptNorms.append(np.asarray(columns[j]).reshape((ntso, -1)))



        # Predict optimums of new tasks
        for k in range(self.problem.DP):
            K = GPy.kern.RBF(input_dim=self.problem.DI)
            M = GPy.models.GPRegression(INorms, PSoptNorms[k], K)
            # M.optimize_restarts(num_restarts = 10, robust=True, verbose=False, parallel=False, num_processes=None, messages="False")
            M.optimize_restarts(num_restarts = kwargs['model_restarts'], robust=True, verbose = kwargs['verbose'], parallel = (kwargs['model_threads'] > 1), num_processes = kwargs['model_threads'], messages = kwargs['verbose'], optimizer = 'lbfgs', start = None, max_iters = kwargs['model_max_iters'], ipython_notebook = False, clear_after_finish = True)
            MSopt.append(M)

        aprxoptsNorm=np.hstack([MSopt[k].predict_noiseless(InewNorms)[0] for k in range(self.problem.DP)])  # the index [0] is the mean value, [1] is the variance
        aprxoptsNorm=np.minimum(aprxoptsNorm,(1-1e-12)*np.ones((ntsn,self.problem.DP)))
        aprxoptsNorm=np.maximum(aprxoptsNorm,(1e-12)*np.ones((ntsn,self.problem.DP)))
        # print('aprxoptsNorm',aprxoptsNorm,type(aprxoptsNorm))
        aprxopts = self.problem.PS.inverse_transform(aprxoptsNorm)
        # print('aprxopts',aprxopts,type(aprxopts),type(aprxopts[0]))


        aprxoptsNormList=[]
        # TnewNormList=[]
        for i in range(ntsn):
            aprxoptsNormList.append([aprxoptsNorm[i,:]])  # this makes sure for each task, there is only one sample parameter set
            # InewNormList.append(InewNorms[i,:])

        t1 = time.time_ns()
        O = self.computer.evaluate_objective(problem = self.problem, I = InewNorms, P =aprxoptsNormList, history_db = self.historydb, options = kwargs)
        t2 = time.time_ns()
        time_fun = time_fun + (t2-t1)/1e9

        #        print(aprxopts)
        #        pickle.dump(aprxopts, open('TLA1.pkl', 'w'))

        t4 = time.time_ns()
        stats['time_total'] = (t4-t3)/1e9
        stats['time_fun'] = time_fun

        return (aprxopts, O, stats)

    def TLA2(): # co-Kriging

        pass


class GPTune_MB(object):

    def __init__(self, tp : TuningProblem, computer : Computer = None, options : Options = None, **kwargs):

        """
        tp: object defining the characteristics of the tuning (See file 'autotuner/autotuner/tuningproblem.py')
        computer     : object specifying the architectural characteristics of the computer to run on (See file 'GPTune/computer.py')
        options      : object defining all the options that will define the behaviour of the tuner (See file 'GPTune/options.py')
        """

        # options contains: bin, bmax, eta
        # Or contains: fidelity map (dictionary from s values to budget level B(s)), eta
        if options['fidelity_map'] == None:
            self.smax = int(np.floor(np.log10(
                options['budget_max']/options['budget_min'])/np.log10(options['budget_base'])))
            self.budgets = [options['budget_max'] /
                            options['budget_base']**x for x in range(self.smax+1)]
            print(f'Using default budgets, smax = {self.smax}, budgets = {self.budgets}.')
        else:
            s_vals = sorted(options['fidelity_map'].keys())
            self.smax = s_vals[-1]
            self.budgets = [options['fidelity_map'].get(key) for key in s_vals]
            print(f'Using user-provided budgets, smax = {self.smax}, budgets = {self.budgets}.')

        parameter_space = tp.parameter_space
        output_space = tp.output_space
        objectives = tp.objective
        constraints = tp.constraints
        constants = tp.constants

        """ insert "budget" as the first dimension of the input space """
        inputs = [Real(options['budget_min']-1e-12,
                       options['budget_max'], transform="normalize", name="budget")]

        for n,p in zip(tp.input_space.dimension_names,tp.input_space.dimensions):
            if (isinstance(p, Real)):
                inputs.append(Real(p.bounds[0], p.bounds[1], transform="normalize", name=n))
            elif (isinstance(p, Integer)):
                inputs.append(Integer(p.bounds[0], p.bounds[1], transform="normalize", name=n))
            elif (isinstance(p, Categorical)):
                inputs.append(Categoricalnorm (list(p.bounds), transform="onehot", name=n))
            else:
                raise Exception("Unknown parameter type")

        print('inputs = ', inputs)
        input_space = Space(inputs)

        self.tp = TuningProblem(input_space, parameter_space,output_space, objectives, constraints, None, constants=constants)
        self.computer = computer
        self.options  = options
        self.data     = Data(tp)


    def MB_LCM(self, NS=None, Igiven=None, Pdefault=None, **kwargs):
        """
        Igiven		 : a list of tasks 
        NS			 : number of samples in the highest budget arm
        Pdefault     : assuming there is a default parameter configuration among all tasks
        """

        np.set_printoptions(suppress=False, precision=4)
        print('\n\n\n------Starting MB_LCM (multi-arm bandit with LCM) with %d loops for task' % (NS), Igiven)

        stats = {
            "time_total": 0,
            "time_sample_init": 0,
            "time_fun": 0,
            "time_search": 0,
            "time_model": 0
        }
        time_fun=0
        time_sample_init=0
        time_search=0
        time_model=0

        # self.NSs = [int(self.options['budget_max']/x*NS) for x in self.budgets] # so that the highest fidelity has NS samples
        self.NSs = [int((self.smax+1)/(s+1))*self.options['budget_base']**s for s in range(self.smax+1)] # consistent with hyperband
        NSs1 = [0] * len(self.NSs)
        info = [[x, y] for x, y in zip(self.budgets, self.NSs)]
        print('total samples:', info)
                
        self.data.I = Igiven
        # self.data.D = [] 
        self.data.P = []
        self.data.O = []
        
        data = Data(self.tp)   # data contains all data in the extended task space returned by MLA

        for Nloop in range(NS):
            data1 = Data(self.tp)  # for each loop, data1 will have all data for all arms sampled by MLA (excluding SH samples)
            data1.I = []
            data1.P = []
            data1.O = []
            data1.D = []
            for s in range(len(self.budgets)):  # loop over the budget levels
                budget = self.budgets[s]
                ns = self.NSs[s] 
                ntotal = NSs1[s] + int(ns)
                NSs1[s] = NSs1[s] + ns
                print(f"Bracket s = {s}, budget = {budget}, ns = {ns}")
                newtasks = []
                if(self.data.D is not None):
                    data.D = []
                for s1 in range(s, len(self.budgets)):
                    for t in range(len(Igiven)):
                        budget1 = self.budgets[s1]
                        tmp = [budget1]+Igiven[t]
                        newtasks.append(tmp)
                        if(self.data.D is not None):
                            if(len(self.data.D)>0):
                                data.D.append(self.data.D[t])
                if s == 0 and Nloop == 0:
                    all_subtasks = copy.deepcopy(newtasks)

                data.I = newtasks
                if(Pdefault is not None):
                    data.P = [[Pdefault]] * len(newtasks) # as gt.MLA will load available database, Pdefault is effective only if database is empty   
                
                # print("Calling MLA: \ndata.I", data.I, "\ndata.P", data.P, "\ndata.O", data.O)
                # print(f"NS={ntotal}, Igiven={newtasks}, NI={len(newtasks)}, NS1={min(self.NSs)}")
                gt = GPTune(self.tp, computer=self.computer,
                            data=data, options=self.options)
                (data, _, stats0) = gt.MLA(NS=ntotal, Igiven=newtasks, NI=len(newtasks), NS1=min(self.NSs))
                data.P = [x[NSs1[s]-ns:NSs1[s]] for x in data.P]   # data has all data (including those from database), only takes ns ones newly generated by MLA 
                data.O = [x[NSs1[s]-ns:NSs1[s]] for x in data.O]
                data1.I += data.I[0:len(Igiven)]     # data1 has data generated by MLA, but only in the current arm
                data1.P += data.P[0:len(Igiven)]
                data1.O += data.O[0:len(Igiven)]
                if(data.D is not None):
                    data1.D += data.D[0:len(Igiven)]
                data.I=[]
                data.P=[]
                data.O=[]
                if(data.D is not None):
                    data.D=None
                # merge new results to history
                
                stats['time_total'] += stats0['time_total']
                stats['time_fun'] += stats0['time_fun']
                stats['time_model'] += stats0['time_model']
                stats['time_search'] += stats0['time_search']
                stats['time_sample_init'] += stats0['time_sample_init']
                # print(f'At the end of bracket {s}')
                # print('Current data1:')
                # print('data1.I: ', data1.I)
                # print('data1.P: ', data1.P)
                # print('data1.O: ', data1.O)
                # print("data1.D = ", data1.D)
                
            if Nloop == 0:
                self.data.P = data1.P[0:len(Igiven)]  # the first 0:len(Igiven) tasks of data1 are highest budget
                self.data.O = data1.O[0:len(Igiven)] 
            else:
                # self.data.P = [np.concatenate((self.data.P[i], data1.P[0:len(Igiven)][i])) for i in range(len(self.data.P))]
                self.data.P = [self.data.P[i] + data1.P[i] for i in range(len(self.data.P))]
                self.data.O = [np.concatenate((self.data.O[i], data1.O[i])) for i in range(len(self.data.O))]
            
            print("Finish multi-arm initial evaluation")
            # print('data.I: ', data.I)
            # print('data.P: ', data.P)
            # print('data.O: ', data.O)
            # print("data.D = ", data.D)
            print('self.data.P = ', self.data.P)
            print('self.data.O = ', self.data.O)
            
            print('\n\n\n------Start SH run on each arm, except the first highest fidelity arm')
            options1 = copy.deepcopy(self.options)
            kwargs.update(options1)
            for s in range(1, len(self.budgets)):
                print('\nArm s=%d'%(s))
                budget = self.budgets[s]
                ns = self.NSs[s]
                # print(f'\n\n\nArm {s}, Initial budget = {budget}, number of total samples = {ns}')
                ratio = int(ns/self.options['budget_base'])
                # print(f'Current s = {s}, budget = {budget}, ns = {ns}')
                idx = s*len(Igiven) 
                temp_I = data1.I[idx:idx+len(Igiven)]
                # print(f'Tasks: ', temp_I)
                temp_O = list(map(np.squeeze, data1.O[idx:idx+len(Igiven)]))
                # temp_P = list(map(np.array, data1.P[idx:idx+len(Igiven)]))
                temp_P = data1.P[idx:idx+len(Igiven)]
                for ri in range(s):
                    idx_sort = list(map(np.argsort, temp_O))
                    temp_O = [y[x[:ratio]] for (x, y) in zip(idx_sort, temp_O)]
                    temp_P = [[y[_i] for _i in x[:ratio]] for (x, y) in zip(idx_sort, temp_P)]
                    # temp_P = [x.tolist() for x in temp_P]
                    # budget *= self.options['budget_base'] # lift budget level
                    budget = self.budgets[s-ri-1]
                    for subtasks in temp_I:
                        subtasks[0] = budget
                    newdata = Data(problem=self.tp, I=temp_I, P=temp_P)
                    gt = GPTune(self.tp, computer=self.computer, data=newdata, options=self.options)
                    
                    
                    gt.history_db.load_history_func_eval(newdata, gt.problem, temp_I)


                    t1 = time.time_ns()
                    done=0
                    if(gt.data.P is not None):
                        if(len(gt.data.P[0])>=NSs1[s-ri-1]+ratio): # the evaluations have been done before 
                            done=1
                    if(done==0):
                        print(f'Evaluating top {ratio} by MLA with budget = {budget}')
                        newdata = Data(problem=self.tp, I=temp_I, P=temp_P)
                        gt = GPTune(self.tp, computer=self.computer, data=newdata, options=self.options)
                        tmp=[]
                        for x in gt.data.P:
                            if(len(x)>0):
                                xNorm = gt.problem.PS.transform(x)
                                tmp.append(xNorm)
                            else:
                                tmp.append(np.empty( shape=(0, gt.problem.DP) ))
                        gt.data.P=tmp
                        # print('gaga',gt.data.P[0])
                        gt.data.I = gt.problem.IS.transform(gt.data.I)
                        newdata.O = gt.computer.evaluate_objective(gt.problem, gt.data.I, gt.data.P, gt.data.D, gt.history_db, options = kwargs)
                        newdata.P = [gt.problem.PS.inverse_transform(x) for x in newdata.P]

                    else:
                        # print('done: what do you do')
                        newdata.I=gt.data.I
                        newdata.P=gt.data.P
                        newdata.O=gt.data.O
                        newdata.P = [elem[NSs1[s-ri-1]:NSs1[s-ri-1]+ratio] for elem in newdata.P]
                        newdata.O = [elem[NSs1[s-ri-1]:NSs1[s-ri-1]+ratio,:] for elem in newdata.O]

                    t2 = time.time_ns()
                    timefun=(t2-t1)/1e9
                    NSs1[s-ri-1] = NSs1[s-ri-1] + ratio
                    
                    temp_O = list(map(np.squeeze, newdata.O))
                    # temp_P = list(map(np.array, newdata.P))
                    temp_P = newdata.P
                    ratio = int(ratio/self.options['budget_base'])
                    stats['time_fun'] += timefun
                    stats['time_total'] += timefun
                # print(newdata.I,'jia')
                # print(newdata.P,'dia')
                newdata.I = Igiven
                # print('newdata before merge')
                # print('newdata.P = ', newdata.P)
                # print('self.data.P = ', self.data.P)
                # self.data.merge(newdata) # this would make self.data.P a list of numpy array
                self.data.P = [x + y for x, y in zip(self.data.P, newdata.P)]
                self.data.O = [np.concatenate((self.data.O[i], newdata.O[i])) for i in range(len(self.data.O))]
                
                # print('Data updated: ')
                # print('self.data.P = ', self.data.P)
                # print('self.data.O = ', self.data.O)
            
            print('Updated self.data after all SH runs')
            print('self.data.P = ', self.data.P)
            print('self.data.O = ', self.data.O)
            
            # data1_hist collects data1 from all loops (it excludes SH samples since it's used to locate arm and sample index of the optimal sample)
            data1.I = all_subtasks # change budgets back to initial values
            if Nloop == 0:
                data1_hist = copy.deepcopy(data1)             
            else:
                data1_hist.merge(data1)
                data1_hist.P = [elem.tolist() for elem in data1_hist.P]
                
            Nloop += 1
            print(f"Finish one loop, next Nloop = {Nloop}")
            print('data1_hist.I = ', data1_hist.I)    
            print('data1_hist.P = ', data1_hist.P)
            print('data1_hist.O = ', data1_hist.O)
            
        return (copy.deepcopy(self.data), stats, data1_hist)

#### Wrapper Functions

def LoadSurrogateModelData(meta_path=None, meta_dict=None, tuning_configuration:dict = None):

    meta_data = {}

    if meta_path != None:
        if os.path.exists(meta_path):
            try:
                with open(meta_path) as f_in:
                    meta_data.update(json.load(f_in))
            except:
                print ("[Error] not able to get model load configuration from path")

    if meta_dict != None:
        try:
            meta_data.update(meta_dict)
        except:
            print ("[Error] not able to get model load configuration from dict")

    input_space_given = meta_data["input_space"]
    parameter_space_given = meta_data["parameter_space"]
    output_space_given = meta_data["output_space"]
    task_parameters_given = meta_data["task_parameters"]
    print ("task_parameters_given: ", task_parameters_given)

    if "modeler" in meta_data:
        modeler = meta_data['modeler']
    else:
        modeler = 'Model_LCM'

    historydb = HistoryDB()
    model_data = historydb.load_surrogate_model_meta_data(
            task_parameters_given,
            tuning_configuration,
            input_space_given,
            parameter_space_given,
            output_space_given,
            0,
            modeler)

    print ("MODEL DATA: ", model_data)

    return (model_data)

def CreateGPTuneFromModelData(model_data):

    input_space_info = model_data["input_space"]
    parameter_space_info = model_data["parameter_space"]
    output_space_info = model_data["output_space"]

    input_space_arr = []
    for input_space_info in model_data["input_space"]:
        name_ = input_space_info["name"]
        type_ = input_space_info["type"]
        transformer_ = input_space_info["transformer"]

        if type_ == "int" or type_ == "Int" or type_ == "Integer" or type_ == "integer":
            lower_bound_ = input_space_info["lower_bound"]
            upper_bound_ = input_space_info["upper_bound"]
            input_space = Integer(lower_bound_, upper_bound_, transform=transformer_, name=name_)
            input_space_arr.append(input_space)
        elif type_ == "real" or type_ == "Real" or type_ == "float" or type_ == "Float":
            lower_bound_ = input_space_info["lower_bound"]
            upper_bound_ = input_space_info["upper_bound"]
            input_space = Real(lower_bound_, upper_bound_, transform=transformer_, name=name_)
            input_space_arr.append(input_space)
        elif type_ == "categorical" or type_ == "Categorical" or type_ == "category" or type_ == "Category":
            categories = input_space_info["categories"]
            input_space = Categoricalnorm(categories, transform=transformer_, name=name_)
            input_space_arr.append(input_space)
    IS = Space(input_space_arr)

    parameter_space_arr = []
    for parameter_space_info in model_data["parameter_space"]:
        name_ = parameter_space_info["name"]
        type_ = parameter_space_info["type"]
        transformer_ = parameter_space_info["transformer"]

        if type_ == "int" or type_ == "Int" or type_ == "Integer" or type_ == "integer":
            lower_bound_ = parameter_space_info["lower_bound"]
            upper_bound_ = parameter_space_info["upper_bound"]
            parameter_space = Integer(lower_bound_, upper_bound_, transform=transformer_, name=name_)
            parameter_space_arr.append(parameter_space)
        elif type_ == "real" or type_ == "Real" or type_ == "float" or type_ == "Float":
            lower_bound_ = parameter_space_info["lower_bound"]
            upper_bound_ = parameter_space_info["upper_bound"]
            parameter_space = Real(lower_bound_, upper_bound_, transform=transformer_, name=name_)
            parameter_space_arr.append(parameter_space)
        elif type_ == "categorical" or type_ == "Categorical" or type_ == "category" or type_ == "Category":
            categories = parameter_space_info["categories"]
            parameter_space = Categoricalnorm(categories, transform=transformer_, name=name_)
            parameter_space_arr.append(parameter_space)
    PS = Space(parameter_space_arr)

    output_space_arr = []
    for output_space_info in model_data["output_space"]:
        name_ = output_space_info["name"]
        type_ = output_space_info["type"]
        transformer_ = output_space_info["transformer"]

        if type_ == "int" or type_ == "Int" or type_ == "Integer" or type_ == "integer":
            lower_bound_ = output_space_info["lower_bound"]
            upper_bound_ = output_space_info["upper_bound"]
            output_space = Integer(lower_bound_, upper_bound_, transform=transformer_, name=name_)
            output_space_arr.append(output_space)
        elif type_ == "real" or type_ == "Real" or type_ == "float" or type_ == "Float":
            lower_bound_ = output_space_info["lower_bound"]
            upper_bound_ = output_space_info["upper_bound"]
            output_space = Real(lower_bound_, upper_bound_, transform=transformer_, name=name_)
            output_space_arr.append(output_space)
        elif type_ == "categorical" or type_ == "Categorical" or type_ == "category" or type_ == "Category":
            categories = output_space_info["categories"]
            output_space = Category(categories, transform=transformer_, name=name_)
            output_space_arr.append(output_space)
    OS = Space(output_space_arr)

    problem = TuningProblem(IS, PS, OS, objective=None, constraints=None, models=None, constants=None)
    computer = Computer(nodes=1, cores=2) # number of nodes/cores is not actually used when reproducing only surrogate models
    data = Data(problem)
    options = Options()
    if "modeler" in model_data:
        options['model_class'] = model_data['modeler']
    else:
        options['model_class'] = 'Model_LCM'
    gt = GPTune(problem, computer=computer, data=data, options=options)

    return (gt)

def LoadSurrogateModelFunction(meta_path="./.gptune/model.json", meta_dict=None, tuning_configuration:dict = None):

    model_data = LoadSurrogateModelData(meta_path, meta_dict, tuning_configuration)
    gt = CreateGPTuneFromModelData(model_data)
    (models, model_function) = gt.LoadSurrogateModel(model_data = model_data)

    return (model_function)

def SensitivityAnalysis(model_data:dict=None, task_parameters=None, num_samples:int=1000):

    gt = CreateGPTuneFromModelData(model_data)
    (models, model_function) = gt.LoadSurrogateModel(model_data = model_data)

    from SALib.sample import saltelli
    from SALib.analyze import sobol

    num_vars = len(model_data["parameter_space"])
    parameter_names = []
    parameter_bounds = []
    parameter_types = []
    categorical_parameter_values = {}
    for parameter_info in model_data["parameter_space"]:
        parameter_name = parameter_info["name"]
        parameter_type = parameter_info["type"]

        parameter_names.append(parameter_name)
        parameter_types.append(parameter_type)

        if parameter_type == "int" or parameter_type == "real":
            lower_bound = parameter_info["lower_bound"]
            upper_bound = parameter_info["upper_bound"]
            parameter_bounds.append([lower_bound, upper_bound])
        elif parameter_type == "categorical":
            lower_bound = 0
            upper_bound = len(parameter_info["categories"])
            parameter_bounds.append([lower_bound, upper_bound])
            categorical_parameter_values[parameter_name] = parameter_info["categories"]
    print ("Categorical_parameter_values: ", categorical_parameter_values)

    parameter_categories = []
    problem = {
            'num_vars': num_vars,
            'names': parameter_names,
            'bounds': parameter_bounds,
            }

    # Generate new samples for sensitivity analysis from the fitted surrogate model.
    parameter_values = saltelli.sample(problem, num_samples)

    Y = []
    for i in range(len(parameter_values)):
        model_input = {}
        for j in range(num_vars):
            if parameter_types[j] == "int" or parameter_types[j] == "real":
                model_input[parameter_names[j]] = int(parameter_values[i][j])
            elif parameter_types[j] == "categorical":
                model_input[parameter_names[j]] = categorical_parameter_values[parameter_names[j]][int(parameter_values[i][j])]

        num_task_parameters = len(model_data["input_space"])
        for j in range(num_task_parameters):
            model_input[model_data["input_space"][j]["name"]] = task_parameters[j]
        #print (model_input)

        output_name = model_data["output_space"][0]["name"]

        Y.append(model_function(model_input)[output_name][0][0])

    print (Y)

    # Sensitivity analysis based on the samples drawn from the fitted surrogate model.
    Si = sobol.analyze(problem, np.array(Y), print_to_console=True)
    #print (Si)
    return Si

def GetSurrogateModelConfigurations(meta_path="./.gptune/model.json", meta_dict=None):

    meta_data = {}

    if meta_path != None:
        if os.path.exists(meta_path):
            try:
                with open(meta_path) as f_in:
                    meta_data.update(json.load(f_in))
            except:
                print ("[Error] not able to get model load configuration from path")

    if meta_dict != None:
        try:
            meta_data.update(meta_dict)
        except:
            print ("[Error] not able to get model load configuration from dict")

    loadable_machine_configurations = meta_data["loadable_machine_configurations"]
    loadable_software_configurations = meta_data["loadable_software_configurations"]

    input_space_given = meta_data["input_space"]
    parameter_space_given = meta_data["parameter_space"]
    output_space_given = meta_data["output_space"]
    task_parameters_given = meta_data["task_parameters"]

    if "modeler" in meta_data:
        modeler = meta_data['modeler']
    else:
        modeler = 'Model_LCM'

    historydb = HistoryDB()
    (model_configurations) = historydb.load_surrogate_model_configurations(
            task_parameters_given,
            input_space_given,
            parameter_space_given,
            output_space_given,
            loadable_machine_configurations,
            loadable_software_configurations,
            0,
            modeler)

    return model_configurations




#! /usr/bin/env python

# GPTune Copyright (c) 2019, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S.Dept. of Energy) and the University of
# California, Berkeley.  All rights reserved.
#
# If you have questions about your rights to use or distribute this software,
# please contact Berkeley Lab's Intellectual Property Office at IPO@lbl.gov.
#
# NOTICE. This Software was developed under funding from the U.S. Department
# of Energy and the U.S. Government consequently retains certain rights.
# As such, the U.S. Government has been granted for itself and others acting
# on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in
# the Software to reproduce, distribute copies to the public, prepare
# derivative works, and perform publicly and display publicly, and to permit
# other to do so.
#

import numpy as np
from problem import Problem
from data import Data
from database import HistoryDB
from typing import Collection, Callable
import mpi4py
# from mpi4py import MPI
import os
import sys
import concurrent
from concurrent import futures

from pathlib import Path
import importlib
import inspect

class Computer(object):

    def __init__(self, nodes : int = 1, cores : int = 1, hosts : Collection = None):

        self.nodes = nodes
        self.cores = cores
        self.hosts = hosts
        if (hosts != None and nodes != len(hosts)):
            raise Exception('The number of elements in "hosts" does not match with the number of "nodes"')

    def evaluate_constraints(self, problem, point : Collection, inputs_only : bool = False, **kwargs):  # point is in the original spaces

#       kwargs['constraints_evaluation_parallelism']

        # points can be either a dict or a list of dicts on which to iterate
        if(problem.constants is not None):
            point.update(problem.constants)
        cond = True
        for (cstname, cst) in problem.constraints.items():
            if (isinstance(cst, str)):
                try:
                    # {} has to be the global argument to eval
                    # and point the local one, otherwise,
                    # point will be corrupted / updated by eval
                    cond = eval(cst, {}, point)
                except Exception as inst:
                    if (inputs_only and isinstance(inst, NameError)):
                        pass
                    else:
                        raise Exception(f"Unexpected exception '{inst}' was raised while evaluating constraint '{cstname}'. Correct this constraint before calling the tuner again.")
            else:
                try:
                    if(hasattr(problem, 'driverabspath')): # differentiate between Problem and TuningProblem 
                        if(problem.driverabspath is not None):
                            modulename = Path(problem.driverabspath).stem  # get the driver name excluding all directories and extensions
                            sys.path.append(problem.driverabspath) # add path to sys
                            module = importlib.import_module(modulename) # import driver name as a module
                            cst = getattr(module, cstname)
                        else:
                            raise Exception('the driverabspath is required for the constraints')        
                    kwargs2 = {}
                    sig = inspect.signature(cst)
                    for varname in point:
                        if (varname in sig.parameters):
                            kwargs2[varname] = point[varname]              
                    cond = cst(**kwargs2)
                except Exception as inst:
                    if (isinstance(inst, TypeError)):
                        lst = inst.__str__().split()
                        if (len(lst) >= 5 and lst[1] == 'missing' and lst[3] == 'required' and lst[4] == 'positional'):
                            pass
                        else:
                            raise Exception(f"Unexpected exception '{inst}' was raised while evaluating constraint '{cstname}'. Correct this constraint before calling the tuner again.")
                    else:
                        raise Exception(f"Unexpected exception '{inst}' was raised while evaluating constraint '{cstname}'. Correct this constraint before calling the tuner again.")
            if (not cond):
                break

        return cond


    def evaluate_objective(self, problem : Problem, I : np.ndarray = None, P : Collection[np.ndarray] = None, D: Collection[dict] = None, history_db : HistoryDB = None, options: dict=None):  # P and I are in the normalized space
        O = []
        for i in range(len(I)):
            T2 = I[i]
            P2 = P[i]
            if D is not None:
                D2 = D[i]
            else:
                D2 = None
            if(options['RCI_mode']==False):    
                O2 = self.evaluate_objective_onetask(problem=problem, i_am_manager=True, T2=T2, P2=P2, D2=D2, history_db=history_db, options=options)

                tmp = np.array(O2).reshape((len(O2), problem.DO))
                O.append(tmp.astype(np.double))   #YL: convert single, double or int to double types

            else:
                tmp = np.empty( shape=(len(P2), problem.DO))
                tmp[:] = np.NaN
                O.append(tmp.astype(np.double))   #YL: NaN indicates that the evaluation data is needed by GPTune

                if history_db is not None:
                    history_db.store_func_eval(problem = problem,\
                            task_parameter = I[i], \
                            tuning_parameter = P[i],\
                            evaluation_result = tmp,\
                            evaluation_detail = tmp)

        if(options['RCI_mode']==True):
            print('RCI: GPTune returns\n')
            exit()

        return O

    def evaluate_objective_onetask(self, problem : Problem, pids : Collection[int] = None, i_am_manager : bool = True, T2 : np.ndarray=None, P2 : np.ndarray=None, D2 : dict=None, history_db : HistoryDB=None, options:dict=None):  # T2 and P2 are in the normalized space

        I_orig = problem.IS.inverse_transform(np.array(T2, ndmin=2))[0]

        if(problem.driverabspath is not None and options['distributed_memory_parallelism']):
            modulename = Path(problem.driverabspath).stem  # get the driver name excluding all directories and extensions
            sys.path.append(problem.driverabspath) # add path to sys
            module = importlib.import_module(modulename) # import driver name as a module
            # func = getattr(module, funcName)
        else:
            module =problem

        O2=[]
        kwargst = {problem.IS[k].name: I_orig[k] for k in range(problem.DI)}

        if (pids is None):
            pids = list(range(len(P2)))

        if (options['distributed_memory_parallelism'] and options['objective_evaluation_parallelism'] and i_am_manager):
            from mpi4py import MPI
            if(problem.driverabspath is None):
                raise Exception('objective_evaluation_parallelism and distributed_memory_parallelism require passing driverabspath to GPTune')

            nproc = min(options['objective_multisample_processes'],len(P2))
            mpi_comm = self.spawn(__file__, nproc, nthreads=1, kwargs=options)
            kwargs_tmp = options
            if "mpi_comm" in kwargs_tmp:
                del kwargs_tmp["mpi_comm"]   # mpi_comm is not picklable
            _ = mpi_comm.bcast((self, problem,P2, D2, I_orig, pids, kwargs_tmp), root=mpi4py.MPI.ROOT)

            tmpdata = mpi_comm.gather(None, root=mpi4py.MPI.ROOT)
            mpi_comm.Disconnect()

            # reordering is needed as tmpdata[p] stores p, p+nproc, p+2nproc, ...
            Otmp=[]
            offset=[0] * (nproc+1)
            for p in range(int(nproc)):
                Otmp = Otmp + tmpdata[p]
                offset[p+1]=offset[p]+len(tmpdata[p])
            for it in range(len(tmpdata[0])):
                for p in range(int(nproc)):
                    if(len(O2)<len(P2)):
                        O2.append(Otmp[offset[p]+it])

            # TODO: HistoryDB function evaluation store

        elif (options['shared_memory_parallelism'] and options['objective_evaluation_parallelism']):
            with concurrent.futures.ThreadPoolExecutor(max_workers = options['objective_multisample_threads']) as executor:
                def fun(pid):
                    x = P2[pid]
                    x_orig = problem.PS.inverse_transform(np.array(x, ndmin=2))[0]
                    kwargs = {problem.PS[k].name: x_orig[k] for k in range(problem.DP)}
                    kwargs.update(kwargst)
                    kwargs.update(D2)
                    if(problem.constants is not None):
                        kwargs.update(problem.constants)
                    # print(kwargs)
                    return module.objectives(kwargs)
                O2 = list(executor.map(fun, pids, timeout=None, chunksize=1))

                # TODO: HistoryDB function evaluation store
        else:

            for j in pids:
                x = P2[j]
                x_orig = problem.PS.inverse_transform(np.array(x, ndmin=2))[0]
                kwargs = {problem.PS[k].name: x_orig[k] for k in range(problem.DP)}
                kwargs.update(kwargst)
                if(problem.constants is not None):
                    kwargs.update(problem.constants)
                if D2 is not None:
                    kwargs.update(D2)
                o = module.objectives(kwargs)
                # print('kwargs',kwargs,'o',o)

                o_eval = []

                if type(o) == type({}): # predicted by model
                    source = o["source"]
                    o_eval = [o[problem.OS[k].name] for k in range(len(problem.OS))]
                else: # type(o) == type([]): # list
                    source = "measure"
                    for i in range(len(o)):
                        if type(o[i]) == type([]):
                            o_eval.append(np.average(o[i]))
                        else:
                            o_eval.append(o[i])

                if history_db is not None:
                    history_db.store_func_eval(problem = problem,\
                            task_parameter = T2, \
                            tuning_parameter = [P2[j]],\
                            evaluation_result = [o_eval], \
                            evaluation_detail = [o], \
                            source = source)
                            #np.array(O2_).reshape((len(O2_), problem.DO)), \

                O2.append(o_eval)

        return O2


    def spawn(self, executable, nproc, nthreads, npernode=None, args=None, kwargs=None):
        from mpi4py import MPI
        print('exec', executable, 'args', args, 'nproc', nproc)

        npernodes=npernode
        if(npernode is None):
            npernodes=self.cores

        info = mpi4py.MPI.Info.Create()
#        info.Set("add-hostfile", "slurm.hosts")
#        info.Set("host", "slurm.hosts")
        info.Set('env', 'OMP_NUM_THREADS=%d\n' %(nthreads))
        info.Set('npernode','%d'%(npernodes))  # YL: npernode is deprecated in openmpi 4.0, but no other parameter (e.g. 'map-by') works


        comm = mpi4py.MPI.COMM_SELF.Spawn(sys.executable, args=executable, maxprocs=nproc,info=info)#, info=mpi_info).Merge()# process_rank = comm.Get_rank()
        # process_rank = comm.Get_rank()
        # process_count = comm.Get_size()
        # process_host = mpi4py.MPI.Get_processor_name()
        # print('manager',process_rank, process_count, process_host)
        return comm


if __name__ == '__main__':
    from mpi4py import MPI

    def objectives(point):
        print('this is a dummy definition')
        return point
    def models(point):
        print('this is a dummy definition')
        return point

    mpi_comm = mpi4py.MPI.Comm.Get_parent()
    mpi_rank = mpi_comm.Get_rank()
    mpi_size = mpi_comm.Get_size()
    (computer, problem,P2, D2, I_orig, pids, kwargs) = mpi_comm.bcast(None, root=0)
    pids_loc = pids[mpi_rank:len(pids):mpi_size]
    tmpdata = computer.evaluate_objective_onetask(problem, pids_loc, False, I_orig, P2, D2, kwargs)
    res = mpi_comm.gather(tmpdata, root=0)
    mpi_comm.Disconnect()
